from .__version__ import __version__

# hack that seems to prevent unexpected crashes
import os
import sys
import pandas as pd
import numpy as np
import torch

from .predict import load_cached_model, predict_scores, predict_best_score, predict_masked, predict_sequence_embedding, predict_residue_embedding


import os
import numpy as np
from sapiens.roberta import RoBERTaSeq2Seq
# register loss function
# noinspection PyUnresolvedReferences
from sapiens.smooth_masked_lm import SmoothMaskedLmLoss
import torch

MODEL_VERSIONS = ['v1']

CACHED_MODELS = {}


def load_cached_model(model_dir, checkpoint_name, cache=True):
    model_key = (model_dir, checkpoint_name)
    if not cache or model_key not in CACHED_MODELS:
        CACHED_MODELS[model_key] = RoBERTaSeq2Seq.load(model_dir, checkpoint_name)
    return CACHED_MODELS[model_key]


def get_model_path(model_version):
    src_root = os.path.dirname(__file__)
    return os.path.join(src_root, 'models', model_version)


def predict_scores(seq, chain_type, model_version='latest', return_all_hiddens=False):
    """Predict Sapiens residue scores for each position in a given sequence

    @param seq: Antibody variable region sequence
    @param chain_type: Chain type (H = heavy, L = light). Same model is used for kappa and lambda.
    @param model_version: Model instance or version of the model: latest, v1, etc
    @param return_all_hiddens: Return a tuple with (probs, weights_dict)
    @return: Pandas DataFrame with one row for each position and one column for each residue type
    """
    assert '/' not in model_version
    if model_version == 'latest':
        model_version = MODEL_VERSIONS[-1]
    model_dir = get_model_path(model_version)
    if chain_type == 'H':
        checkpoint_name = 'checkpoint_vh.pt'
    elif chain_type in ['K', 'L']:
        checkpoint_name = 'checkpoint_vl.pt'
    else:
        raise ValueError(f'Unknown chain type {chain_type}')
    model = load_cached_model(model_dir, checkpoint_name)
    return model.predict_scores(seq, return_all_hiddens=return_all_hiddens)


def predict_best_score(seq, chain_type, model_version='latest'):
    scores = predict_scores(seq, chain_type, model_version=model_version)
    return ''.join(scores.idxmax(axis=1).values)


def predict_masked(seq, chain_type, model_version='latest'):
    best_score_seq = predict_best_score(seq, chain_type, model_version=model_version)
    return ''.join([b if a in '*X' else a for a, b in zip(seq, best_score_seq)])


def predict_residue_embedding(seq, chain_type, layer=None):
    pred, extra = predict_scores(seq, chain_type=chain_type, model_version='latest', return_all_hiddens=True)
    embed = extra['inner_states']
    if layer is not None:
        return embed[layer][1:-1,0,:].cpu().numpy()
    return np.array([e[1:-1,0,:].cpu().numpy() for e in embed])


def predict_sequence_embedding(seq, chain_type, layer=None):
    return predict_residue_embedding(seq, chain_type, layer=layer).mean(axis=-2)



import math
import torch
from fairseq import metrics, modules, utils
from fairseq.criterions import register_criterion, FairseqCriterion
from fairseq.criterions.label_smoothed_cross_entropy import label_smoothed_nll_loss

@register_criterion('smooth_masked_lm')
class SmoothMaskedLmLoss(FairseqCriterion):
    """
    Label-smoothed implementation for the loss used in masked language model (MLM) training.
    """

    def __init__(self, task, label_smoothing):
        super().__init__(task)
        self.eps = label_smoothing

    @staticmethod
    def add_args(parser):
        """Add criterion-specific arguments to the parser."""
        # fmt: off
        parser.add_argument('--label-smoothing', default=0., type=float, metavar='D',
                            help='epsilon for label smoothing, 0 means no label smoothing')
        # fmt: on

    def forward(self, model, sample, reduce=True):
        """Compute the loss for the given sample.

        Returns a tuple with three elements:
        1) the loss
        2) the sample size, which is used as the denominator for the gradient
        3) logging outputs to display while training
        """
        # compute MLM loss
        masked_tokens = sample['target'].ne(self.padding_idx)

        # Rare: when all tokens are masked, project all tokens.
        # We use torch.where to avoid device-to-host transfers,
        # except on CPU where torch.where is not well supported
        # (see github.com/pytorch/pytorch/issues/26247).
        if masked_tokens.device == torch.device('cpu'):
            if not masked_tokens.any():
                masked_tokens.fill_(True)
        else:
            masked_tokens = torch.where(
                masked_tokens.any(),
                masked_tokens,
                masked_tokens.new([True]),
            )

        net_output = model(**sample['net_input'], masked_tokens=masked_tokens)
        logits, _ = net_output
        targets = model.get_targets(sample, [logits])
        targets = targets[masked_tokens]

        lprobs = model.get_normalized_probs(net_output, log_probs=True)

        loss, nll_loss = label_smoothed_nll_loss(
            lprobs.view(-1, lprobs.size(-1)),
            targets.view(-1, 1),
            self.eps,
            reduce=reduce,
            ignore_index=self.padding_idx
        )

        sample_size = masked_tokens.int().sum()
        logging_output = {
            'loss': loss.data,
            'nll_loss': nll_loss.data,
            'ntokens': sample['ntokens'],
            'nsentences': sample['nsentences'],
            'sample_size': sample_size,
        }
        return loss, sample_size, logging_output

    @staticmethod
    def reduce_metrics(logging_outputs) -> None:
        """Aggregate logging outputs from data parallel training."""
        loss_sum = sum(log.get('loss', 0) for log in logging_outputs)
        nll_loss_sum = sum(log.get('nll_loss', 0) for log in logging_outputs)
        sample_size = sum(log.get('sample_size', 0) for log in logging_outputs)

        metrics.log_scalar('loss', loss_sum / sample_size / math.log(2), sample_size, round=3)
        metrics.log_scalar('nll_loss', nll_loss_sum / sample_size / math.log(2), sample_size, round=3)
        metrics.log_derived('ppl', lambda meters: utils.get_perplexity(meters['nll_loss'].avg))

    @staticmethod
    def logging_outputs_can_be_summed() -> bool:
        """
        Whether the logging outputs returned by `forward` can be summed
        across workers prior to calling `reduce_metrics`. Setting this
        to True will improves distributed training speed.
        """
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import os
from io import open

about = {}
# Read version number from sapiens.__version__.py (see PEP 396)
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'sapiens', '__version__.py'), encoding='utf-8') as f:
    exec(f.read(), about)

# Read contents of readme file into string
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='sapiens',
    version=about['__version__'],
    description='Sapiens: Human antibody language model based on BERT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='David Prihoda',
    packages=find_packages(include=['sapiens.*', 'sapiens']),
    author_email='david.prihoda@gmail.com',
    license='MIT',
    python_requires=">=3.7",
    install_requires=[
        'pandas',
        'fairseq'
    ],
    keywords='sapiens, antibody humanization, bert, biophi',
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    include_package_data=True,
    package_data={'': ['*.pt', '*.txt']},
    url='https://github.com/Merck/Sapiens'
)



from fairseq.models import register_model_architecture
from fairseq.models.roberta import RobertaModel, RobertaHubInterface
from fairseq.models.roberta.model import base_architecture
from fairseq.tasks.sentence_prediction import SentencePredictionTask
from fairseq.tasks.masked_lm import MaskedLMTask
import fairseq
import fairseq.utils
import torch
import pandas as pd
from dataclasses import dataclass
import os


def seq2tokens(seq, dictionary, add_bos):
    assert ' ' not in seq, f'Expected regular sequence without space separators, found space in "{seq}"'
    sentence = ('<s> ' if add_bos else '') + seq2sentence(seq).replace('X', '<mask>').replace('*', '<mask>') + ' </s>'
    return dictionary.encode_line(sentence, append_eos=False, add_if_not_exist=False).long()


def seq2sentence(seq):
    """Convert sequence to string of space-separated letters"""
    return ' '.join(list(seq))


@dataclass
class RoBERTa:
    interface: RobertaHubInterface

    @classmethod
    def load(cls, model_dir, checkpoint_name='checkpoint.pt') -> 'RoBERTa':
        full_model_dir = os.path.abspath(model_dir)
        if os.path.exists(full_model_dir):
            interface = RobertaModel.from_pretrained(
                full_model_dir,
                checkpoint_name,
                full_model_dir,
                bpe=None
            )
            interface.eval()
            return cls(interface=interface)
        else:
            raise FileNotFoundError(f'Model not found: {full_model_dir}')

    def predict_scores(self, seq):
        raise NotImplementedError()


@dataclass
class RoBERTaSeq2Seq(RoBERTa):

    def predict_scores(self, seq, remove_special=True, return_all_hiddens=False):
        """
        Get model output probabilities for all positions in a single sequence (slow, consider batching for large inputs)
        :param seq: Input sequence (str or Bio.Seq)
        :param remove_special: remove special dictionary tokens from predicted probability matrix
        :param return_all_hiddens: return attention from all layers (used with return_attention)
        :return: 2D dataframe of output token probabilities (columns) for each input position (rows)
        """
        add_bos = self._is_adding_bos()
        tokens = self._source_seq_tokens(seq, add_bos=add_bos)
        with torch.no_grad():
            try:
                logits, extra = self.interface.model(
                    tokens.unsqueeze(0),
                    features_only=False,
                    return_all_hiddens=return_all_hiddens
                )
            except IndexError:
                raise ValueError(f'Sequence is too long ({len(seq)} residues), make sure to only include the variable region: "{seq}"')
            pred = logits.softmax(dim=-1)[0]
            # remove EOS token
            pred = pred[:-1, :]
            if add_bos:
                # remove BOS token
                pred = pred[1:, :]
        pred = pd.DataFrame(pred.numpy(), columns=self.interface.task.target_dictionary.symbols)
        if remove_special:
            pred.drop(['<s>', '<pad>', '</s>', '<unk>', '<mask>'], axis=1, inplace=True)
        if return_all_hiddens:
            return pred, extra
        else:
            return pred

    def _is_adding_bos(self):
        if isinstance(self.interface.task, SentencePredictionTask):
            return True
        elif isinstance(self.interface.task, MaskedLMTask):
            return True
        return ValueError(f'Unknown task: {type(self.interface.task)}')

    def _source_seq_tokens(self, seq, add_bos):
        return seq2tokens(seq, dictionary=self.interface.task.source_dictionary, add_bos=add_bos)

    def _target_seq_tokens(self, seq, add_bos):
        return seq2tokens(seq, dictionary=self.interface.task.target_dictionary, add_bos=add_bos)


@register_model_architecture('roberta', 'roberta_small')
def roberta_small_architecture(args):
    args.encoder_layers = getattr(args, 'encoder_layers', 4)
    args.encoder_embed_dim = getattr(args, 'encoder_embed_dim', 128)
    args.encoder_ffn_embed_dim = getattr(args, 'encoder_ffn_embed_dim', 256)
    args.encoder_attention_heads = getattr(args, 'encoder_attention_heads', 8)
    base_architecture(args)
